/* Copyright (c) 2019-2022 Adrian Lopez

   This software is provided 'as-is', without any express or implied
   warranty. In no event will the authors be held liable for any damages
   arising from the use of this software.

   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must not
      claim that you wrote the original software. If you use this software
      in a product, an acknowledgment in the product documentation would be
      appreciated but is not required.
   2. Altered source versions must be plainly marked as such, and must not be
      misrepresented as being the original software.
   3. This notice may not be removed or altered from any source distribution. */

#ifndef MBSTOWCS_ESCAPE_INVALID_H
#define MBSTOWCS_ESCAPE_INVALID_H

#include <wchar.h>

size_t mbstowcs_escape_invalid(wchar_t *dest, const char *src, size_t n);

#endif